#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "ogg/ogg.h"
#include "vorbis/codec.h"

#define MAXSIZE 512
#include <stdint.h>
#include "iodefine.h"
#include "ivects.h"
#include "tty.h"
#include "cprintf.h"
#include "timer.h"
#include "sdram.h"

#define BUFFER_SIZE	128
#define USE_DAC		0

#define LED0	PORT0.DR.BIT.B2
#define LED1	PORT0.DR.BIT.B3
#define LED2	PORT0.DR.BIT.B5
#define LED3	PORT3.DR.BIT.B4

/*--------------------------------------------------------------*/

static char * sbase = 0;

void *
sbrk (int adj)
{
  char *prev_sbase;
  if (sbase == 0)
    sbase = sdram_base ();
  prev_sbase = sbase;
  sbase += adj;
  cprintf("sbrk(%l) = %l..%l\n", adj, prev_sbase, sbase);
  return prev_sbase;
}

/*--------------------------------------------------------------*/
/* Generic Buffer Functions					*/

typedef struct {
  uint16_t left;
  uint16_t right;
} BufferEntry[BUFFER_SIZE];

typedef struct {
  int get_ptr, put_ptr;
  volatile int count;
  volatile BufferEntry data;
} Buffer;

static inline void
buffer_put (Buffer *b, uint16_t left, uint16_t right)
{
  int i = b->put_ptr;
  b->put_ptr = (b->put_ptr + 1) % BUFFER_SIZE;
  b->count ++;
  b->data[i].left = left;
  b->data[i].right = right;
}

static inline void
buffer_get (Buffer *b, uint16_t *left, uint16_t *right)
{
  int i = b->get_ptr;
  b->get_ptr = (b->get_ptr + 1) % BUFFER_SIZE;
  b->count --;
  *left = b->data[i].left;
  *right = b->data[i].right;
}

static inline int
buffer_full (Buffer *b)
{
  return b->count == BUFFER_SIZE;
}

static inline int
buffer_empty (Buffer *b)
{
  return b->count == 0;
}

Buffer adc_buffer;
#if USE_DAC
Buffer dac_buffer;
#endif

/*--------------------------------------------------------------*/
/* ADC Functions						*/

static int p6 = 0;

static void __attribute__((interrupt))
adc_handler (void)
{
  uint16_t left, right;
  left  = S12AD.ADDRA;
  right = S12AD.ADDRB;
  if (!buffer_full (&adc_buffer))
    buffer_put (&adc_buffer, left, right);
}

void
adc_init (void)
{
  MSTP(S12AD) = 0;
  S12AD.ADANS.WORD = 0x0003; /* scan channels 0 and 1 */
  S12AD.ADADS.WORD = 0x0003;
  S12AD.ADADC.BYTE = 3;
  S12AD.ADCSR.BYTE = 0x1e;
  S12AD.ADCER.BIT.ADRFMT = 1; /* left aligned data */
  S12AD.ADSTRGR.BYTE = 0x3; /* TRGAN_0 */

  _vectors[VECT(S12AD,ADI)] = adc_handler;
  IEN(S12AD,ADI) = 1;
  IPR(S12AD,ADI) = 5;

  MSTP(MTU0) = 0;
  MTU1.TCR.BYTE = 0x20; /* TGRA, PCLK/1 = 48 MHz */
  MTU1.TIER.BIT.TTGE = 1; /* ADC trigger enabled */
  MTU1.TGRA = 1088; /* should be 44100 KHz */
}

void
adc_start (void)
{
  MTUA.TSTR.BIT.CST1 = 1; /* start timer */
}

void
adc_stop (void)
{
  MTUA.TSTR.BIT.CST1 = 0; /* start timer */
}

void
adc_rate (int samples_per_second)
{
  MTU1.TGRA = 48000000 / samples_per_second;
}

/*--------------------------------------------------------------*/
/* DAC Functions						*/
#if USE_DAC

static void __attribute__((interrupt))
mtu2_handler ()
{
  uint16_t left, right;

  if (! buffer_empty (&dac_buffer))
    {
      buffer_get (&dac_buffer, &left, &right);
      DA.DADR0 = left;
      DA.DADR1 = right;
    }
}

void
dac_init (void)
{
  MSTP(DA) = 0;
  DA.DACR.BYTE = 0xe0;
  DA.DADPR.BYTE = 0x80; /* left-aligned */
  PORT0.DDR.BIT.B3 = 1;
  PORT0.DDR.BIT.B5 = 1;

  MSTP(MTU0) = 0;
  _vectors[VECT(MTU2,TGIA2)] = mtu2_handler;
  IEN(MTU2,TGIA2) = 1;
  IPR(MTU2,TGIA2) = 5;

  MTU2.TCR.BYTE = 0x20; /* TGRA, PCLK/1 = 48 MHz */
  MTU2.TIER.BYTE = 0x01; /* interrupt on TGRA */
  MTU2.TGRA = 1088; /* should be 44100 KHz */
  MTUA.TSTR.BIT.CST2 = 1; /* start timer */
}

void
dac_rate (int samples_per_second)
{
  MTU2.TGRA = 48000000 / samples_per_second;
}

#endif
/*--------------------------------------------------------------*/

int
main(void)
{
  uint16_t min, max;
  unsigned long last_print = 0;

  SYSTEM.SCKCR.LONG = 0x00010100;

  PORT0.DDR.BIT.B2 = 1;
  PORT0.DDR.BIT.B3 = 1;
  PORT0.DDR.BIT.B5 = 1;
  PORT3.DDR.BIT.B4 = 1;

  LED0 = LED1 = LED2 = LED3 = 1;
  LED3 = 0;

  sdram_init ();

  timer_init ();
  LED0 = 0;
  LED3 = 1;
  tty_init ();
  LED1 = 0;
  tty_puts("\nOGG TEST\n");
  LED2 = 0;

  setbuf (stdout, NULL);
  setbuf (stderr, NULL);

  printf("printf works.\n");

#if USE_DAC
  dac_init ();
  dac_rate (44100);
#endif

  usb_init ();
  adc_init ();
  adc_rate (44100);

  LED3 = 0;

  //adc_start ();

  char *buf;
  int r, n;

  ogg_sync_state oy;
  ogg_page og;
  ogg_stream_state os;
  ogg_packet op;
  int saw_vorbis = 0;

  vorbis_info vi;
  vorbis_comment vc;
  vorbis_dsp_state v;
  vorbis_block vb;

  float **pcm;


  ogg_sync_init (&oy);
  vorbis_info_init (&vi);
  vorbis_comment_init (&vc);

  while (1)
    {
      while (ogg_sync_pageout (&oy, &og) != 1)
	{
	  buf = ogg_sync_buffer (&oy, MAXSIZE);
	  r = usb_read (buf, MAXSIZE);
	  printf("\033[32mREAD: %d\033[0m\n", r);
	  if (r <= 0)
	    goto end_of_file;
	  ogg_sync_wrote (&oy, r);
	}

      printf("OGG Page %d:", ogg_page_serialno (&og));
      for (r=0; r<og.header_len; r++)
	printf(" %02x", og.header[r]);
      printf("\n");
      printf("OGG Page %d:", ogg_page_serialno (&og));
      for (r=0; r<og.header_len; r++)
	printf(" %02c", isgraph (og.header[r]) ? og.header[r] : ' ');
      printf("\n");

      n = og.body_len < 32 ? og.body_len : 32;
      printf("OGG Body:", ogg_page_serialno (&og));
      for (r=0; r<n; r++)
	printf(" %02x", og.body[r]);
      printf("\n");
      printf("OGG Body:", ogg_page_serialno (&og));
      for (r=0; r<n; r++)
	printf(" %02c", isgraph (og.body[r]) ? og.body[r] : ' ');
      printf("\n");

      if (og.body[0] == 1 && memcmp (og.body+1, "vorbis", 6) == 0)
	{
	  ogg_stream_init (&os, ogg_page_serialno (&og));
	  saw_vorbis = 1;
	  printf("\033[34mVORBIS\033[0m\n");
	}
      if (!saw_vorbis)
	continue;

      if (ogg_stream_pagein (&os, &og) != 0)
	{
	  printf("\033[31mNot my page\033[0m\n");
	  continue;
	}

    more_packetout:
      r = ogg_stream_packetout (&os, &op);
      switch (r)
	{
	case -1: printf("sync/gap\n"); break;
	case  0: printf("need more data\n"); break;
	case  1:
	  {
	    printf("my packet!\n");

	    n = op.bytes < 32 ? op.bytes : 32;
	    printf("\033[36mOGG Packet:");
	    for (r=0; r<n; r++)
	      printf(" %02x", op.packet[r]);
	    printf("\n");
	    printf("OGG Body:");
	    for (r=0; r<n; r++)
	      printf(" %02c", isgraph (op.packet[r]) ? op.packet[r] : ' ');
	    printf("\033[0m\n");

	    if (saw_vorbis <= 3)
	      {
		saw_vorbis ++;
		vorbis_synthesis_headerin (&vi, &vc, &op);

		if (saw_vorbis == 4)
		  {
		    vorbis_synthesis_init (&v, &vi);
		    vorbis_block_init (&v, &vb);
		    printf("\033[33mBitstream: %d channels, %d sps\n", vi.channels, vi.rate);
		  }
	      }
	    else
	      {
		vorbis_synthesis (&vb, &op);
		vorbis_synthesis_blockin (&v, &vb);

		r = vorbis_synthesis_pcmout (&v, &pcm);
		printf("\033[35mPCM: %d\033[0m\n", r);
		for (n=0; n<r; n++)
		  {
		    int left, right;

		    if (pcm[0][n] < -1 || pcm[0][n] > 1)
		      printf("[%d] %10f %10f\n", n, pcm[0][n], pcm[1][n]);
		    left = pcm[0][n] * 32768;
		    right = pcm[1][n] * 32768;
#if USE_DAC
		    buffer_put (&dac_buffer, left, right);
#endif
		  }
		vorbis_synthesis_read (&v, r);
	      }

	  }
	  goto more_packetout;
	}
    }

 end_of_file:
  printf("done!\n");
  for (;;);
}
