#include <stdio.h>
#include <string.h>
#include "ports.h"
#include "cprintf.h"

/* LEDS are configured as follows, looking at the stick with the usb
   cable DOWN:

		RED[7:0]  RED[13:8]
		P20..P26  P27,PA0..PA5  LOW = ON
   COM0 PD0
   COM1
   COM2  .
   COM3
   COM4  .

   COM5  .
   COM6
   COM7 PD7
   COM8 PE0  LOW = ON
   COM9 PE1
*/

static unsigned short pixels[10][14];

unsigned int
rand(void)
{
  static unsigned long long next = 1;
  /* This multiplier was obtained from Knuth, D.E., "The Art of
     Computer Programming," Vol 2, Seminumerical Algorithms, Third
     Edition, Addison-Wesley, 1998, p. 106 (line 26) & p. 108 */
  next = next * 6364136223846793005LL + 1;
  return (int)(next >> 32);
}

/*
  P2.0	CB3
  P2.1	CA3
  P2.2	CC3
  P2.3	CD3
  P2.4	CB4
  P2.5	CA4
  P2.6	CA5
  P2.7	CB5

  PA.0	CA6
  PA.1	CB6
  PA.2	CC6
  PA.3	CD6
  PA.4	CA7
  PA.5	CB7
  PA.6	CA8
  PA.7	CB8
*/

static volatile int timer = 0;
static volatile int volume;

static void __attribute__((interrupt))
tpu0_handler ()
{
  static int row = 0;
  int com;
  timer ++;

  volume = ADC[1].addrd;
  ADC[1].adcsr = 0x23; /* start conversion */

  if (timer & 1)
    {
      com = ~(1 << row);
      DR[2] = DR[10] = ~0;
      DR[13] = com;
      DR[14] = com >> 8;
#if 0
      DR[2] = pixels[row];
      DR[10] = pixels[row] >> 8;
#else
      TPU[3].tgrb = pixels[row][0];
      TPU[3].tgra = pixels[row][1];
      TPU[3].tgrc = pixels[row][2];
      TPU[3].tgrd = pixels[row][3];
      TPU[4].tgrb = pixels[row][4];
      TPU[4].tgra = pixels[row][5];
      TPU[5].tgra = pixels[row][6];
      TPU[5].tgrb = pixels[row][7];
      TPU[7].tgrb = pixels[row][8];
      TPU[7].tgra = pixels[row][9];
      TPU[7].tgrc = pixels[row][10];
      TPU[7].tgrd = pixels[row][11];
      TPU[8].tgrb = pixels[row][12];
      TPU[8].tgra = pixels[row][13];
#endif
      row = (row + 1) % 10;
    }
}

extern int16_t audio[];
static volatile int16_t *audio_data = NULL;
extern int16_t audio_end[];

static void __attribute__((interrupt))
tpu1_handler ()
{
  if (audio_data)
    {
      DADR0 = ((*audio_data ++ * volume) >> 16) + 0x8000;
      if (audio_data == audio_end)
	audio_data = NULL;
    }
}

static void
wait_ms (int ms)
{
  int now = timer;
  while ((int)(timer - now) < ms)
    ;
}

char banner_message[] = "DJ was here!\n";

/* 0,0 is the lower left, 13,9 is the upper right */
static int
pixel (c, r, v)
{
  if (r < 0 || r > 9 || c < 0 || c > 13)
    return;
  pixels[r][c] = v;
}

void
cls()
{
  memset (pixels, 0, sizeof(pixels));
}

extern unsigned char image[];
#define IMG_OFS 0x36
#define IMG_WIDTH 300
#define IMG_SCALE 5
#define IMG_SX	31
#define IMG_SY	0
#define IMG_EX	230
#define IMG_EY	303

#define img(x,y) image[IMG_OFS + (x) + (y)*IMG_WIDTH]

main()
{
  int i, j, r, c;

  /* Power up the peripherals.  */
  MSTPCRA = 0;
  MSTPCRB = 0;
  MSTPCRC = 0;

  set_system_clock_speed ();

  DR[2] = 0xfe;
  DR[10] = 0xfe;
  DR[13] = 0xfe;
  DR[14] = 0xfe;

  DDR[6] = 0xc0; /* DAC outputs */
  DDR[2] = 0xff;
  DDR[10] = 0xff;
  DDR[13] = 0xff;
  DDR[14] = 0xff;

  /* DAC - we use DAC0 */
  DACR = 0xff;
  DADPR = 0x80;

  /* Timer */
  _vectors[IVEC_TPU0_TGI0A] = tpu0_handler;
  IER0D |= 0x01;
  IPR4C = 5;

  _vectors[IVEC_TPU1_TGI1A] = tpu1_handler;
  IER0D |= 0x80;
  IPR4E = 5;

  TPU[0].tcr = 0x20; /* TGRA, PCLK = 50 MHz */
  TPU[0].tier = 0x01; /* interrupt on TGRA */
  TPU[0].tgra = 50000; /* should be 1 mS */

  TPU[1].tcr = 0x20; /* TGRA, PCLK = 50 MHz */
  TPU[1].tier = 0x01; /* interrupt on TGRA */
  TPU[1].tgra = 1134; /* about 22,050 Hz */

  for (i=2; i<=9; i++)
    {
      if (i == 6)
	continue;
      TPU[i].tcr = 0x60; /* cleared by sync clear */
      TPU[i].tmdr = 0x03; /* simple PWM2 */
      TPU[i].tiorh = 0x22;
      TPU[i].tiorl = 0x22;
      TPU[i].tgra = 64;
      TPU[i].tgrb = 128;
      TPU[i].tgrc = 192;
      TPU[i].tgrd = 255;
    }
  TPU[2].tcr = 0x20; /* TGRA */
  TPU[2].tgra = 256;
  TPU[9].tcr = 0x20; /* TGRA */
  TPU[9].tgra = 256;

  TPU[2].tcnt = 0;
  TPU[9].tcnt = 0;

  TSYRA = 0x3c; /* Synchronize TPU 2, 3, 4, 5 */
  TSYRB = 0x07; /* Synchronize TPU 6, 7, 8 */

  TSTRA |= 0x3f; /* start timer */
  TSTRB |= 0x07; /* start timer */


  ADC[1].adcr = 0x08; /* software trigger, PCLK/8, single mode */
  ADC[1].addpr = 0x80;

  setup_interrupts ();

  tty_init ();
  wait_ms (10);
  tty_puts(banner_message);

  i = 1;
  if (*(unsigned char *)&i == 1)
    tty_puts("little endian\n");
  else
    tty_puts("big endian\n");

  while (1)
    {

      cls();

      for (i=0; i<30; i++)
	{
	  pixel(i/2, 0, 30);
	  pixel(i/6, i/6, 100);
	  pixel(0, i/3, 255);
	  wait_ms (10);
	}
      wait_ms (500);

      for (i=0; i<12; i++)
	for (j=0; j<=i; j++)
	  {
	    int p = 255;
	    pixel(i-j, j, p);
	    pixel(13-(i-j), 9-j, p);
	    wait_ms (5);
	  }

      wait_ms (500);
      {
	int mix[10][14];
	for (r=0; r<10; r++)
	  for (c=0; c<14; c++)
	    mix[r][c] = rand() % 768;

	for (i=0; i<1024; i++)
	  {
	    for (r=0; r<10; r++)
	      for (c=0; c<14; c++)
		{
		  if (pixels[r][c] > 0
		      && i > mix[r][c])
		    pixels[r][c] --;
		}
	    wait_ms (1);
	  }
      }
      wait_ms (500);

      audio_data = audio;
      {
	double f, fi;
	for (f=0; f<1.0; f+= 1.0/375.0)
	  {
	    fi = 1.0 - f;
	    int x = IMG_SX * fi + IMG_EX * f;
	    int y = IMG_SY * fi + IMG_EY * f;
	    for (r=0; r<10; r++)
	      for (c=0; c<14; c++)
		pixels [9-r][c] = img(x + c * IMG_SCALE, y + r * IMG_SCALE);
	    wait_ms (8);
	  }
      }
      wait_ms (500);
    }

  i = 1;
  while (1)
    {
      int p0, p2, p3, p8;

      for( r = 0; r<10; r++ )
	{
	  for( c=0; c<14; c++)
	    {
	      
	      for (j=800; j; j--)
		{
		  for(p0=0; p0<100; p0++)
		    asm("");
		if( j%2 )
		  {
		    pixel(r,c);
		  }
		else
		  {
		    pixel(9-r, 13-c);
		  }
		}
	      //i = (unsigned int) rand() % 140;


	      //r = i / 10;
	      //c = i % 10;

	      //pixel (r, c);
	    }
	}
    }
}
