#include <stdio.h>
#include <stdint.h>
#include "iodefine.h"
#include "ivects.h"
#include "tty.h"
#include "timer.h"
#include "usb.h"
#include "cprintf.h"
#include "critsec.h"

#define DEBUG 0

#define LED0	PORT0.DR.BIT.B2
#define LED1	PORT0.DR.BIT.B3
#define LED2	PORT0.DR.BIT.B5
#define LED3	PORT3.DR.BIT.B4

/* Reads:  OUT endpoint 1, pipe 1, D1FIFO */
/* Writes: IN  endpoint 2, pipe 2, D0FIFO */

#define REP	1
#define WEP	2
#define RPIPE	1
#define WPIPE	2
#define BULK_PACKET_SIZE 64

#define RXBUFSZ 128
static char rxbuf[RXBUFSZ];
static volatile int rxput, rxget, rxcount;

#define TXBUFSZ 8192
static char txbuf[TXBUFSZ];
static volatile int txput, txget, txcount;

static inline void
portc ()
{
  PORTC.DR.BIT.B5 = (txcount >= TXBUFSZ-1) ? 1 : 0;
  PORTC.DR.BIT.B4 = (txcount >= BULK_PACKET_SIZE) ? 1 : 0;
  PORTC.DR.BIT.B3 = (txcount <= 1) ? 1 : 0;
}

#define PID_NAK 	0
#define PID_BUF 	1
#define PID_STALL_1	2
#define PID_STALL_2	3

volatile int usb_connected = 0;

/* These are the USB descriptors we'll be returning.  The values and
   structures are standard USB and can be found in the USB spec
   (www.usb.org).  */

static const USB_device_descriptor my_device_d = {
  sizeof(USB_device_descriptor),
  DESCRIPTOR_DEVICE,
  0x0110,
  0,
  0,
  0,
  64, /* max packet size for control pipe */
  MY_VENDOR_ID,
  0x0025, /* "DJ" */
  0x0101, /* device release number */
  1,
  0,
  0,
  1
};

static USB_configuration_descriptor my_configuration_d = {
  sizeof(USB_configuration_descriptor),
  DESCRIPTOR_CONFIGURATION,
  sizeof(USB_configuration_descriptor), /* total length */
  1, /* num interfaces */
  1, /* configuration value */
  0,
  CONFIG_ATTR_RESERVED_SET | CONFIG_ATTR_SELF_POWERED,
  10, /* 20 mA */
};

static const USB_interface_descriptor my_interface_d = {
  sizeof(USB_interface_descriptor),
  DESCRIPTOR_INTERFACE,
  0,
  0,
  2, /* num endpoints */
  0xff, /* class */
  0xff, /* subclass */
  0xff, /* protocol */
  0
};

static const USB_endpoint_descriptor my_endpoint1in_d = {
  sizeof(USB_endpoint_descriptor),
  DESCRIPTOR_ENDPOINT,
  1 | EP_OUT,
  EP_BULK,
  64,
  1
};

static const USB_endpoint_descriptor my_endpoint2out_d = {
  sizeof(USB_endpoint_descriptor),
  DESCRIPTOR_ENDPOINT,
  2 | EP_IN,
  EP_BULK,
  64,
  1
};

static const USB_lang_descriptor my_lang_d = {
  4,
  DESCRIPTOR_STRING,
  0x0409,
};

/* Note that USB strings are 16-bit Unicode, and not all compilers can
   do 16-bit strings (Linux, for example, uses 32-bit wchar_t).  */
static const USB_string_descriptor my_string1_d = {
  2 + 2*13,
  DESCRIPTOR_STRING,
  /* requires gcc -std=gnuc99 */
  u"DJ Delorie RX"
};

#define pending USB0.INTSTS0.BIT

static char buf[200];

static void
usb0_ctrt_handler (void)
{
  int first;
  pending.CTRT = 0;

#if DEBUG
  PORT6.DR.BYTE = 2;
  siprintf(buf, "%3d.%03d CTRT state %d addr %04x req %02x.%02x val %04x idx %04x len %04x : ",
	   timer_ms / 1000, timer_ms % 1000,
	   USB0.INTSTS0.BIT.CTSQ,
	   USB0.USBADDR.WORD,
	   USB0.USBREQ.BIT.BREQUEST,
	   USB0.USBREQ.BIT.BMREQUESTTYPE,
	   USB0.USBVAL,
	   USB0.USBINDX,
	   USB0.USBLENG);
  PORT6.DR.BYTE = 4;
  tty_puts (buf);
#endif

  PORT6.DR.BYTE = 8;

  switch (USB0.INTSTS0.BIT.CTSQ)
    {
    case 0: /* idle or setup */
      break;

    case 1: /* control read data */
    case 3: /* control write data */
      switch (USB0.USBREQ.WORD)
	{
	case 0x0080: /* status */
	case 0x0081:
	case 0x0082:
	  tty_putc('s');
	  usb_write_pipe (0, (uint8_t *)"\0\0", 2);
	  break;

	case 0x0680: /* request descriptor */
	  {
	    switch (USB0.USBVAL)
	      {
	      case 0x0100: /* device */
		tty_putc('d');
		usb_write_pipe (0, (uint8_t *)&my_device_d, sizeof(my_device_d));
		break;

	      case 0x0200: /* configuration */
		tty_putc('c');
		/* This is called twice if needed.  if there isn't
		   enough room, it's called again with a bigger
		   buffer.  The device must send all interface and
		   endpoint descriptors along with the configuration
		   descriptor.  */
		my_configuration_d.wTotalLength =
		  sizeof (my_configuration_d)
		  + sizeof(my_interface_d)
		  + sizeof(my_endpoint1in_d)
		  + sizeof(my_endpoint2out_d);
		first = (USB0.USBLENG < my_configuration_d.wTotalLength) ? 1 : 0;
		usb_write_start (0);
		if (first)
		  {
		    int len = sizeof(my_configuration_d);
		    if (len > USB0.USBLENG)
		      len = USB0.USBLENG;
		    usb_write_data ((uint8_t *)&my_configuration_d, len);
		  }
		else
		  {
		    usb_write_data ((uint8_t *)&my_configuration_d, sizeof(my_configuration_d));
		    usb_write_data ((uint8_t *)&my_interface_d, sizeof(my_interface_d));
		    usb_write_data ((uint8_t *)&my_endpoint1in_d, sizeof(my_endpoint1in_d));
		    usb_write_data ((uint8_t *)&my_endpoint2out_d, sizeof(my_endpoint2out_d));
		  }
		usb_write_done ();
		break;

		/* There's one of these for each string, the LSB is the index.  */
	      case 0x0300:
		tty_putc('E');
		usb_write_pipe (0, &my_lang_d, 4);
		break;
	      case 0x0301:
		tty_putc('V');
		usb_write_pipe (0, &my_string1_d, 2 + 2*13);
		break;

	      default:
		tty_puts(" val? ");
		break;
	      }
	  }
	  break;

	default:
	  tty_puts(" req? ");
	  break;
	}
      break;

    case 5: /* control write no-data-status */
      switch (USB0.USBREQ.WORD) {
      case 0x0900: /* set config */
	/* We only have one config, so we just ACK the request.  */
	usb_setup_ack ();
	USB0.PIPE1CTR.BIT.PID = PID_BUF;
	usb_connected = 1;
	break;
      }

    case 2: /* control read status */
    case 4: /* control write status */
    case 6: /* sequence error */
    case 7: /* doesn't happen */
      break;
    }
}

static int rx_paused = 0;

static void
usb0_brdy_handler (void)
{
  int rv = 0;
  if (USB0.BRDYSTS.BIT.PIPE2BRDY)
    {
      int remaining = BULK_PACKET_SIZE;

#if DEBUG
      cprintf ("BRDY2: %d ", txcount);
#endif

      if (txcount == 0)
	{
	  USB0.CFIFOCTR.BIT.BCLR = 1;
	  USB0.BRDYENB.BIT.PIPE2BRDYE = 0;
	}
      else
	{
	  USB0.D0FIFOSEL.BIT.CURPIPE = WPIPE;
	  while(USB0.D0FIFOCTR.BIT.FRDY == 0){;}
	  while (remaining > 0 && txcount > 0)
	    {
	      int psw;
	      int b = txbuf[txget++];
	      USB0.D0FIFO.BYTE.L = b;
	      txget %= TXBUFSZ;
	      txcount --;
	      portc();
	      remaining --;
	    }
	  USB0.PIPE2CTR.BIT.PID = PID_BUF;
	  if (remaining > 0)
	    USB0.D0FIFOCTR.BIT.BVAL = 1;
	}
    }

  if (USB0.BRDYSTS.BIT.PIPE1BRDY)
    {
      int count;
      uint8_t b;

      USB0.BRDYSTS.BIT.PIPE1BRDY = 0;

      siprintf(buf, "\033[32mB%04x\033[0m", USB0.D1FIFOCTR.WORD);
      tty_puts(buf);
      

      PORT6.DR.BYTE = 2;
      while (USB0.D1FIFOCTR.BIT.FRDY == 0)
	;
      PORT6.DR.BYTE = 4;
      USB0.D1FIFOSEL.BIT.RCNT = 1;


      count = USB0.D1FIFOCTR.BIT.DTLN;

      siprintf(buf, "\033[32mB%04x.%d\033[0m", USB0.BRDYSTS.WORD, count);
      tty_puts(buf);

      while (count > 0)
	{
	  count --;
	  b = USB0.D1FIFO.BYTE.L;
	  rxbuf[rxput++] = b;
	  rxput %= RXBUFSZ;
	  rxcount ++;
	}
      if (rxcount < RXBUFSZ - 64)
	{
	  USB0.PIPE1CTR.BIT.PID = PID_BUF;
	  rx_paused = 0;
	}
      else
	{
	  USB0.PIPE1CTR.BIT.PID = PID_NAK;
	  rx_paused = 1;
	}
    }
}

int
usb_read (unsigned char *buf, int maxlen)
{
  int rv = 0;
#if DEBUG
  siprintf (buf, "usb_read %d\n", maxlen);
  tty_puts (buf);
#endif
  while (rxcount == 0)
    ;
  while (rxcount > 0 && maxlen > 0)
    {
      *buf++ = rxbuf[rxget++];
      rxget %= RXBUFSZ;
      rxcount --;
      rv ++;
      maxlen --;
    }
  if (rxcount < RXBUFSZ - 64 && rx_paused)
    {
      USB0.PIPE1CTR.BIT.PID = PID_BUF;
      rx_paused = 0;
    }
#if 0
  siprintf (buf, "< rxbuf %d\n", rxcount);
  tty_puts (buf);
#endif
  return rv;
}

int
usb_write (unsigned char *buf, int len)
{
  int psw;
  PORTC.DR.BIT.B2 = 1;
  while (len > 0)
    {
      if (txcount >= TXBUFSZ)
	{
	  tty_puts("> txbuf full\n");
	  while (txcount >= TXBUFSZ)
	    USB0.BRDYENB.BIT.PIPE2BRDYE = 1;
	  tty_puts("> txbuf ready\n");
	}
      txbuf[txput++] = *buf++;
      txput %= TXBUFSZ;
      ENTER_CRITICAL_SECTION(psw);
      txcount ++;
      LEAVE_CRITICAL_SECTION(psw);
      portc ();
      len --;
    }
  if (txcount >= BULK_PACKET_SIZE)
    USB0.BRDYENB.BIT.PIPE2BRDYE = 1;
  PORTC.DR.BIT.B2 = 0;
}

static void __attribute__((interrupt))
usb0_handler (void)
{
#if DEBUG
  tty_puts("\033[31m[[\033[0m\n");
#endif
  __builtin_rx_mvtipl (5);
  __builtin_rx_setpsw ('I');
  PORTC.DR.BIT.B6 = 1;

  PORT6.DR.BYTE = 1;

  while (1)
    {
#if DEBUG
      tty_putc('!');
      siprintf(buf, "%04x", USB0.INTSTS0.WORD);
      tty_puts(buf);
#endif
      if (pending.CTRT)
	usb0_ctrt_handler ();
      else if (pending.BRDY)
	usb0_brdy_handler ();
      else
	break;
#if DEBUG
      tty_putc('\n');
#endif
    }

  PORT6.DR.BYTE = 0;
#if DEBUG
  tty_puts("\033[31m]]\033[0m\n");
#endif
  PORTC.DR.BIT.B6 = 0;
}

void
usb_init (void)
{
  usb_connected = 0;

  PORT6.DDR.BYTE = ~0;
  PORT6.DR.BYTE = 0;

  IOPORT.PFKUSB.BYTE = 0x1c; /* Enable, PDHZ, PUPHZ, function mode */

  USB0.SYSCFG.WORD = 0;
  USB0.SYSCFG.BIT.SCKE = 1; /* supply clock */
  USB0.SYSCFG.BIT.USBE = 1; /* enable USB0 */
  USB0.BRDYENB.WORD = 0;

  /* WRITE PIPE - pipe 2, endpoint 2 */
  USB0.PIPESEL.BIT.PIPESEL = WPIPE;
  USB0.PIPE2CTR.WORD = 0;
  while(USB0.PIPE2CTR.BIT.PBUSY)
    ;
  USB0.PIPECFG.WORD = 0;
  USB0.PIPECFG.BIT.EPNUM = WEP;
  USB0.PIPECFG.BIT.DIR = 1;
  USB0.PIPECFG.BIT.BFRE = 0;
  USB0.PIPECFG.BIT.TYPE = 1;
  USB0.PIPEMAXP.BIT.MXPS = BULK_PACKET_SIZE;
  USB0.BRDYENB.BIT.PIPE2BRDYE = 0; /* set this later */

  USB0.PIPE2CTR.BIT.SQCLR = 1;
  USB0.PIPE2CTR.BIT.ACLRM = 1;
  USB0.PIPE2CTR.BIT.ACLRM = 0;
  USB0.PIPE2CTR.BIT.PID = PID_BUF;
  USB0.D0FIFOSEL.BIT.CURPIPE = WPIPE;

  /* READ PIPE - pipe 1, endpoint 1 */
  USB0.PIPESEL.BIT.PIPESEL = RPIPE;
  USB0.PIPE1CTR.WORD = 0;
  while(USB0.PIPE1CTR.BIT.PBUSY)
    ;
  USB0.PIPECFG.WORD = 0;
  USB0.PIPECFG.BIT.EPNUM = REP;
  USB0.PIPECFG.BIT.DIR = 0;
  USB0.PIPECFG.BIT.BFRE = 0;
  USB0.PIPECFG.BIT.TYPE = 1;
  USB0.PIPEMAXP.BIT.MXPS = BULK_PACKET_SIZE;
  USB0.BRDYENB.BIT.PIPE1BRDYE = 1; /* read pipe, always be ready */

  USB0.PIPE1CTR.BIT.ACLRM = 1;
  USB0.PIPE1CTR.BIT.SQCLR = 1;
  USB0.PIPE1CTR.BIT.ATREPM = 0;
  USB0.PIPE1CTR.BIT.PID = PID_BUF;
  USB0.D1FIFOSEL.BIT.CURPIPE = RPIPE;
  USB0.D1FIFOCTR.BIT.BCLR = 1;
  USB0.PIPE1CTR.BIT.ACLRM = 0;

  USB0.PIPESEL.BIT.PIPESEL = 0;

  _vectors[VECT(USB0,USBI0)] = usb0_handler;
  IEN(USB0,USBI0) = 1;
  IPR(USB0,USBI0) = 4;

  USB0.INTENB0.WORD = 0x0800;
  USB0.INTENB1.WORD = 0;
  USB0.INTENB0.BIT.BRDYE = 1;

  USB0.SYSCFG.BIT.DPRPU = 1; /* pullup to say "I'm here" */
}

void
usb_uninit (void)
{
 USB0.SYSCFG.BIT.DPRPU = 0;
}

void
usb_setup_ack (void)
{
  USB0.INTSTS0.BIT.VALID = 0;
  USB0.DCPCTR.BIT.PID = PID_BUF;
  USB0.DCPCTR.BIT.CCPL = 1;
}

void
usb_write_start (int pipe)
{
  USB0.INTSTS0.BIT.VALID = 0;
  USB0.DCPCTR.BIT.PID = PID_BUF;

  USB0.CFIFOSEL.WORD = pipe | 0x4020;
}

void
usb_write_data (const void *vdata, int len)
{
  const uint8_t *data = (uint8_t *)vdata;
  int i;
  tty_putc('B');
  cprintf("%d", len);
  while (USB0.DCPCTR.BIT.BSTS == 0)
    ;
  tty_putc('W');
  for (i=0; i<len; i++)
    USB0.CFIFO.WORD = data[i];
  tty_putc('E');
}

void
usb_write_done ()
{
  USB0.DCPCTR.BIT.CCPL = 1;
  USB0.CFIFOCTR.BIT.BVAL = 1;
}

void
usb_write_more ()
{
  USB0.CFIFOCTR.BIT.BVAL = 1;
}

/* This is a shortcut.  Each start/data/done cycle generates one
 transaction and an ACK, so if you need multiple transactions, use
 start/data/more for the not-last ones.  */
void
usb_write_pipe (int pipe, const void *vdata, int len)
{
  tty_putc('[');
  usb_write_start (pipe);
  usb_write_data (vdata, len);
  usb_write_done ();
  tty_putc(']');
}
